function answer = demo3D23( param )
% demo3D23
%
% Synthesize five 3D images from a microtubule model,
% with sampling mode set to 'sampled' at a density of 25 and
% convolution with point-spread function.
% Results will be seven TIFF files, one each for
% cell boundary, nuclear boundary, nucleoli, mitochondria, lysosomes,
% endosomes, and microtubules, in folder "synthesizedImages/cell1"

% Created: Ivan E. Cao-Berg
%
% Copyright (C) 2014 Murphy Lab
% Lane Center for Computational Biology
% School of Computer Science
% Carnegie Mellon University
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published
% by the Free Software Foundation; either version 2 of the License,
% or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
% General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
% 02110-1301, USA.
%
% For additional information visit http://murphylab.web.cmu.edu or
% send email to murphy@cmu.edu

curr_path = which('demo3D23.m');
curr_path = curr_path(1:end-10);
cd(curr_path);

outputDirectory = pwd;
if nargin == 0
    param = [];
end

param = ml_initparam( param, ...
    struct( 'numberOfSynthesizedImages', 5 ) );
param = ml_initparam( param, ...
    struct( 'seed', 3 ) );

outputDirectory = pwd;
param.targetDirectory = outputDirectory;
param.prefix = 'synthesizedImages';
param.compression = 'lzw';
param.microscope = 'svi';
param.sampling.method = 'sampled';
param.sampling.density = 25;
param.verbose = true;
param.debug = true;

answer = slml2img( {'../../../models/3D/nuc.mat', ...
  '../../../models/3D/centro.mat'...
  '../../../models/3D/tub.mat'}, param );

folder = [ pwd filesep 'temp' filesep 'microtubules' ];
files = dir( [ folder filesep 'microtub*.mat' ] );

disp( 'Postprocessing demo images' );
projections_folder = [ pwd filesep 'projections' ];

if exist( projections_folder )
    rmdir( projections_folder, 's' );
end

mkdir( projections_folder );

for i=1:1:length(files)
    filename = [ folder filesep files(i).name ];
    disp( filename )
    temp = load( filename );
    
    img1 = tif2img( [ pwd filesep 'synthesizedImages' ...
    filesep 'cell' num2str(i) filesep ...
    'nucleus.tif' ] );

    img2 = tif2img( [ pwd filesep 'synthesizedImages' ...
        filesep 'cell' num2str(i) filesep ...
        'cell.tif' ] );
    
    microtubules = zeros( size(img1) );
    
    for j=1:1:length(temp.microtubules.scaled_microtubules)
        scaled_microtubule = temp.microtubules.scaled_microtubules{j};
        for k=1:1:length(scaled_microtubule)
            point = scaled_microtubule(k,:,:);
            microtubules(point(1), point(2), point(3)) = ...
                microtubules(point(1), point(2), point(3)) + 1;
        end
    end
    
    directory = [ pwd filesep 'projections' filesep 'cell' ...
        num2str(i) ];
    if ~exist( directory )
        mkdir( directory );
    end

    img = combine3( img2, img1, microtubules );
    img = im2projection_RGB( img );
    filename = [ directory '.png' ];
    imwrite( img, filename );
    clear filename
end
end

function img = combine( img1, img2 )
img1( find( img1 ~= 0 ) ) = 1;
img2( find( img2 ~= 0 ) ) = 1;

img = img1;
img( find( img2 == 1 ) ) = 2;
end%combine

function img = combine3( img1, img2, img3 )
    img1( find( img1 ~= 0 ) ) = 1;
    img2( find( img2 ~= 0 ) ) = 1;
    img3( find( img3 ~= 0 ) ) = 1;
    
    img = img1;
    img( find( img2 == 1 ) ) = 2;
    img( find( img3 == 1 ) ) = 3;
end%combine
